﻿using UnityEngine;
using UnityEngine.UI;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;
// VAMStory - VAMSDistance v1.0
//
// Check the distance between two things and execute triggers when crossing thresholds

namespace VAMStory
{
    public class VAMSDistance : MVRScript
    {
	    // Static string
	    public static string PLUGIN_PATH;
	    public static string EXCEPTION_PREFIX = "VAMSDistance - Exception caught: ";
	    public static string CAM_OBJ_LABEL = "Player (Main Camera)";
	    
	    // JSONStorables
		protected JSONStorableStringChooser DistanceReferenceObj;
		public JSONStorableFloat DistanceThreshold;
		public JSONStorableFloat DistanceTransitionLowThreshold;

		// Lists
		protected List<string> _ignoredAtoms = new List<String>() {"[CameraRig]","PlayerNavigationPanel","CoreControl"};
		protected List<string> _referenceObjsChoices = new List<string>() {CAM_OBJ_LABEL};
		protected List<Atom> _atomInstances = new List<Atom>();
		
		// State
		protected UIDynamicTextField VAMSDistState;
		
		// Gameobjects, transforms...
		protected Camera _VAMSDistanceCam;
		protected Transform _TRCurrent;
		protected Transform _TRTarget;
		
		// Triggers
		private List<EventTrigger> allTriggers;
		private EventTrigger thresholdCrossedBelow;
		private EventTrigger thresholdCrossedAbove;
		private EventTrigger thresholdTransition;

		// Flags
		protected bool _isInit = false;
		protected bool _aboveTriggerFired = false;
		protected bool _belowTriggerFired = false;
		
		// Floats
		protected float _objectDistance = 0f;
		protected float _objectDistanceRemap = 0f;
			
		public override void Init()
        {

			try
            {
	            PLUGIN_PATH = Utils.GetPluginPath(this);

	            // Triggers
	            allTriggers = new List<EventTrigger>();
			
	            thresholdCrossedBelow = new EventTrigger(this, "thresholdCrossedBelow");
	            allTriggers.Add(thresholdCrossedBelow);
	            
	            thresholdCrossedAbove = new EventTrigger(this, "thresholdCrossedAbove");
	            allTriggers.Add(thresholdCrossedAbove);
	            
	            thresholdTransition = new EventTrigger(this, "thresholdTransition");
	            allTriggers.Add(thresholdTransition);

	            thresholdTransition.active = true;
	            
	            // Loading triggers assets
	            SimpleTriggerHandler.LoadAssets();
	            
	            /* *** Temporary UIDynamic *** */
	            UIDynamicTextField tmpTextfield;
	            UIDynamicPopup tmpPopup;
	            UIDynamicSlider tmpSlider;
	            UIDynamicButton tmpButton;
	            UIDynamicToggle tmpToggle;
	            UIDynamic tmpSpacer;
	            
	            /* UI */
                DistanceReferenceObj = new JSONStorableStringChooser("DistanceReferenceObj", _referenceObjsChoices, _referenceObjsChoices.First(), "Reference Object");
                DistanceReferenceObj.setCallbackFunction += (val) => { OnChangeReferenceObject(); };
				CreateScrollablePopup(DistanceReferenceObj, false);

	            DistanceThreshold = new JSONStorableFloat("DistanceThreshold", 3f, 0f, 10f, false);
	            tmpSlider = CreateSlider(DistanceThreshold, false);
	            tmpSlider.label = "Distance threshold";
	            
	            DistanceTransitionLowThreshold = new JSONStorableFloat("DistanceTransitionLowThreshold", 0f, 0f, 10f, false);
	            DistanceTransitionLowThreshold.setCallbackFunction += (val) => { DistanceTransitionLowThreshold.valNoCallback = Mathf.Clamp(val, 0f, DistanceThreshold.val); };
	            tmpSlider = CreateSlider(DistanceTransitionLowThreshold, false);
	            tmpSlider.label = "Transition lower threshold";

	            VAMSDistState = Utils.CreateStaticDescriptionText(this,"Stateinfos","State",false,36,TextAnchor.MiddleLeft,false,false);
	            Utils.SetupTextField(VAMSDistState,230f,false,false);
	            
	            tmpButton = CreateButton("On threshold crossed (below)", false);
	            tmpButton.button.onClick.AddListener(() => { thresholdCrossedBelow.OpenPanelActionStart(); } );

	            tmpButton = CreateButton("On threshold crossed (above)", false);
	            tmpButton.button.onClick.AddListener(() => { thresholdCrossedAbove.OpenPanelActionStart(); } );
	            
	            tmpButton = CreateButton("Threshold transition (from 0 to the threshold value)", false);
	            tmpButton.button.onClick.AddListener(() => { thresholdTransition.OpenPanelTransition(); } );
	            
	            Utils.CreateStaticDescriptionText(this,"VAMSDistanceTitle","<color=#000><size=35><b>VAMSDistance</b></size></color>",true,40,TextAnchor.MiddleRight);
	            string pluginDesc =
		            "<color=#000><size=30>Checks the distance between this object and another. Select the target and set triggers when the thresholds are crossed.\n\n" +
		            "By default the plugin checks the distance between this object and the player (the main camera), but you can change the reference object to compare the distance between this object and any other object in the scene.\n\n" +
		            "The threshold transition trigger allows you to create a smooth transition between <b>Transition lower threshold</b> and <b>threshold</b> value." +
		            "</size></color>";
	            Utils.CreateStaticDescriptionText(this,"Right UI Main Title",pluginDesc,true,590,TextAnchor.MiddleRight);
	            
	            /* ***************************************** */
	            /* **** REGISTERING STORABLES/ACTIONS   **** */
	            /* ***************************************** */
                RegisterStringChooser(DistanceReferenceObj);
	            RegisterFloat(DistanceThreshold);
	            RegisterFloat(DistanceTransitionLowThreshold);
	            
	            // Camera
	            _VAMSDistanceCam = SuperController.singleton.centerCameraTarget.targetCamera;
	            
	            // Transforms
	            _TRCurrent = containingAtom.reParentObject.GetChild(0).transform; // Getting "object"
	            _TRTarget = _VAMSDistanceCam.transform;
	            
	            // Coroutine
	            StartCoroutine("CheckDistance");
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start()
        {
			try
			{
				// Loading atoms references
				_referenceObjsChoices = new List<string>() {CAM_OBJ_LABEL};
				foreach(Atom at in SuperController.singleton.GetAtoms() )
				{
					if (_ignoredAtoms.Contains(at.name)) continue;
					_atomInstances.Add(at);
					_referenceObjsChoices.Add(at.name);
				}

				DistanceReferenceObj.choices = _referenceObjsChoices;
				_isInit = true;
				
				// Initializing ref
				OnChangeReferenceObject();
			}
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
        }
		
		void Update()
		{
			thresholdTransition.TriggerTransition(_objectDistanceRemap);
		}
		
		// **************************
		// Functions
		// **************************
		protected void UpdateState()
		{
			string stateTxt = "";

			stateTxt += "<b>Distance:</b> " + _objectDistance + "\n";
			stateTxt += "<b>Distance (Transition):</b> " + _objectDistanceRemap + "\n";
			stateTxt += "<b>Target:</b> " + (_objectDistance >= DistanceThreshold.val ? "above" : "below") + " threshold";
			VAMSDistState.text = stateTxt;
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		protected void OnChangeReferenceObject()
		{
			if (!_isInit) return;
			if (DistanceReferenceObj.val != CAM_OBJ_LABEL)
			{
				_TRTarget = _atomInstances[_referenceObjsChoices.IndexOf(DistanceReferenceObj.val) - 1].reParentObject.GetChild(0).transform;
			}
			else
			{
				_TRTarget = _VAMSDistanceCam.transform;
			}
		}
		
		// **************************************
		// Coroutines
		// **************************************
		protected IEnumerator CheckDistance()
		{
			// Waiting for URLAudioClip to be ready
			while (true)
			{
				// An update every two frame seems to yield enough accuracy
				yield return new WaitForSeconds(0.03f);

				_objectDistance = Vector3.Distance( _TRCurrent.position, _TRTarget.position );
				_objectDistanceRemap = Mathf.Clamp( Utils.Remap(_objectDistance, DistanceTransitionLowThreshold.val, DistanceThreshold.val, 0f, 1f), 0f, 1f);
				
				// Firing triggers
				if (_objectDistance >= DistanceThreshold.val && _aboveTriggerFired == false)
				{
					_belowTriggerFired = false;
					_aboveTriggerFired = true;
					thresholdCrossedAbove.Trigger();
				} else if (_objectDistance < DistanceThreshold.val && _belowTriggerFired == false)
				{
					_belowTriggerFired = true;
					_aboveTriggerFired = false;
					thresholdCrossedBelow.Trigger();
				}
				UpdateState();
			}
		}

		// **************************
		// OVERRIDES FOR THE TRIGGERS
		// **************************	
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				foreach( EventTrigger et in allTriggers ) {
					jc[et.Name] = et.GetJSON(base.subScenePrefix);
				}
				
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				foreach( EventTrigger et in allTriggers ) {
					et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				}
			}
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		

		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			StopCoroutine("CheckDistance");
		}

		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Threshold Transition";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
		
	}
}